from syntax import *

def eval_aexp(env: dict[str,int], exp: ArithExpr) -> int | None:
    """evaluates an arithmetic expression"""
    match exp:
        case AECst(value):
            return value
        case AEVar(var):
            return env.get(var, None)
        case AEUop(uop, expr):
            v = eval_aexp(env, expr)
            if v is None:
                return None
            match uop:
                case Uop.OPP:
                    return -v
                case _:
                    return None
        case AEBop(bop, left_expr, right_expr):
            left_val = eval_aexp(env, left_expr)
            right_val = eval_aexp(env, right_expr)
            if left_val is None or right_val is None:
                return None
            match bop:
                case Bop.ADD:
                    return left_val + right_val
                case Bop.MUL:
                    return left_val * right_val
                case Bop.DIV:
                    if right_val == 0:
                        return None
                    return left_val // right_val
                case _:
                    return None
        case _:
            assert False

def eval_bexp(env: dict[str,int], exp: BoolExpr) -> bool | None:
    """evaluates a boolean expression"""
    match exp:
        case BEPlain(aexpr):
            val = eval_aexp(env, aexpr)
            if val is None:
                return None
            return val != 0
        case BEEq(left_expr, right_expr):
            left_val = eval_aexp(env, left_expr)
            right_val = eval_aexp(env, right_expr)
            if left_val is None or right_val is None:
                return None
            return left_val == right_val
        case BELeq(left_expr, right_expr):
            left_val = eval_aexp(env, left_expr)
            right_val = eval_aexp(env, right_expr)
            if left_val is None or right_val is None:
                return None
            return left_val <= right_val
        case BENeg(expr):
            val = eval_bexp(env, expr)
            if val is None:
                return None
            return not val
        case _:
            assert False

def eval_instr(env: dict[str,int], instr: Instr) -> dict[str,int] | None:
    """evaluates an instruction"""
    match instr:
        case ISkip():
            return env
        case ISet(var, expr):
            val = eval_aexp(env, expr)
            if val is None:
                return None
            new_env = env.copy()
            new_env[var] = val
            return new_env
        case ISeq(first, second):
            env1 = eval_instr(env, first)
            if env1 is None:
                return None
            return eval_instr(env1, second)
        case ICond(cond, true_branch, false_branch):
            b = eval_bexp(env, cond)
            if b is None:
                return None
            if b:
                return eval_instr(env, true_branch)
            else:
                return eval_instr(env, false_branch)
        case ILoop(cond, body):
            current_env = env
            while True:
                b = eval_bexp(current_env, cond)
                if b is None:
                    return None
                if not b:
                    break
                new_env = eval_instr(current_env, body)
                if new_env is None:
                    return None
                current_env = new_env
            return current_env
        case _:
            assert False